# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


"""
This package contains various command line wrappers to programs used in
pymatgen that do not have Python equivalents.
"""
